use crate::blossom::Blossom;
use crate::dvm::tiktok::TikTokDvm;
use anyhow::Result;
use nostr_sdk::prelude::DataVendingMachineStatus;
use nostr_sdk::{Client, Event, EventBuilder, EventId, Kind, Tag, Timestamp, Url};
use std::collections::HashMap;
use std::future::Future;
use std::pin::Pin;
use std::time::Duration;

mod tiktok;

#[derive(Clone)]
pub struct DVMJobRequest {
    /// The source event
    pub event: Event,
    /// Input data for the job (zero or more inputs)
    pub inputs: Vec<DVMInput>,
    /// Expected output format. Different job request kind defines this more precisely.
    pub output_type: Option<String>,
    /// Optional parameters for the job as key (first argument)/value (second argument).
    /// Different job request kind defines this more precisely. (e.g. [ "param", "lang", "es" ])
    pub params: HashMap<String, String>,
    /// Customer MAY specify a maximum amount (in millisats) they are willing to pay
    pub bid: Option<u64>,
    /// List of relays where Service Providers SHOULD publish responses to
    pub relays: Vec<String>,
}

#[derive(Clone)]
pub enum DVMInput {
    Url {
        url: Url,
        relay: Option<String>,
        marker: Option<String>,
    },
    Event {
        event: EventId,
        relay: Option<String>,
        marker: Option<String>,
    },
    Job {
        event: EventId,
        relay: Option<String>,
        marker: Option<String>,
    },
    Text {
        data: String,
        relay: Option<String>,
        marker: Option<String>,
    },
}

/// Basic DVM handler that accepts a job request
pub trait DVMHandler: Send + Sync {
    fn handle_request(
        &mut self,
        request: DVMJobRequest,
    ) -> Pin<Box<dyn Future<Output = Result<()>> + Send>>;
}

/// Return list of all enabled DVMs
pub fn create_dvms(client: &Client, blossom: &Blossom) -> Vec<Box<dyn DVMHandler>> {
    let ret: Vec<Box<dyn DVMHandler>> =
        vec![Box::new(TikTokDvm::new(client.clone(), blossom.clone()))];
    ret
}

pub fn build_status_for_job(
    req: &DVMJobRequest,
    status: DataVendingMachineStatus,
    extra: Option<&str>,
    content: Option<&str>,
) -> EventBuilder {
    EventBuilder::new(Kind::JobFeedback, content.unwrap_or("")).tags([
        Tag::parse(["status", status.to_string().as_str(), extra.unwrap_or("")]).unwrap(),
        Tag::expiration(Timestamp::now() + Duration::from_secs(30)),
        Tag::event(req.event.id),
        Tag::public_key(req.event.pubkey),
    ])
}
