use crate::blossom::Blossom;
use crate::dvm::{build_status_for_job, DVMHandler, DVMInput, DVMJobRequest};
use anyhow::{bail, Result};
use log::info;
use nostr_sdk::prelude::DataVendingMachineStatus;
use nostr_sdk::{serde_json, Client, EventBuilder, Kind, Tag};
use serde::{Deserialize, Serialize};
use std::env::temp_dir;
use std::fmt::format;
use std::fs::create_dir_all;
use std::future::Future;
use std::path::{Path, PathBuf};
use std::pin::Pin;
use std::time::Duration;
use yt_dlp::fetcher::deps::Libraries;
use yt_dlp::utils::executor::Executor;
use yt_dlp::Youtube;

/// Basic DVM which clones videos from TikTok and posts them as kind 22 shorts
pub struct TikTokDvm {
    client: Client,
    blossom: Blossom,
}

impl TikTokDvm {
    pub fn new(client: Client, blossom: Blossom) -> Self {
        Self { client, blossom }
    }
}

impl DVMHandler for TikTokDvm {
    fn handle_request(
        &mut self,
        request: DVMJobRequest,
    ) -> Pin<Box<dyn Future<Output = anyhow::Result<()>> + Send>> {
        let client = self.client.clone();
        let blossom = self.blossom.clone();
        Box::pin(async move {
            // TODO: process multiple inputs
            let input = if let Some(i) = request.inputs.first() {
                i
            } else {
                bail!("No job inputs found")
            };

            let input = if let DVMInput::Url { url, .. } = input {
                url
            } else {
                bail!("Only URL inputs are accepted")
            };

            if input.host_str() != Some("www.tiktok.com") {
                bail!("Only tiktok urls are accepted");
            }

            // Send starting status
            let starting =
                build_status_for_job(&request, DataVendingMachineStatus::Processing, None, None);
            client.send_event_builder(starting).await?;

            // search for lis
            let tmp_dir = temp_dir().join("tiktok_dvm");
            if !tmp_dir.exists() {
                create_dir_all(&tmp_dir)?;
            }

            let lib_dir = tmp_dir.join("libs");
            let libs = Libraries::new(lib_dir.join("yt-dlp"), lib_dir.join("ffmpeg"));
            libs.install_youtube().await?;

            let dlp = TikTok::new(libs);

            let job_id = request.event.id;
            let output_file = temp_dir().join(&format!("{}.mp4", request.event.id.to_hex()));
            info!("[{}] Starting download of: {}", job_id, input);
            dlp.download_video_from_url(input.as_str(), &output_file)
                .await?;

            info!(
                "[{}] Uploading video from: {}",
                job_id,
                output_file.display()
            );

            let signer = client.signer().await?;
            let res = blossom
                .upload(&output_file, &signer, Some("video/mp4"))
                .await?;

            let info = dlp.get_video_info(input.as_str()).await?;

            let mut imeta = vec!["imeta".to_string()];
            for (k, v) in res.nip94.unwrap() {
                imeta.push(format!("{} {}", k, v));
            }

            let ev = EventBuilder::new(Kind::Custom(22), &info.full_title)
                .tag(Tag::parse(imeta)?)
                .tags([
                    Tag::parse(["title", &info.full_title])?,
                    Tag::parse(["r", input.as_str()])?,
                ]);

            let event_posted = client.send_event_builder(ev).await?;
            let job_data = NoviaVideoData {
                event_id: event_posted.val.to_hex(),
                video: res.sha256,
            };

            let status = build_status_for_job(
                &request,
                DataVendingMachineStatus::Success,
                None,
                Some(&serde_json::to_string(&job_data)?),
            );
            client.send_event_builder(status).await?;

            info!("[{}] Job finished", job_id);
            Ok(())
        })
    }
}

struct TikTok {
    libs: Libraries,
}

impl TikTok {
    pub fn new(libs: Libraries) -> Self {
        Self { libs }
    }

    pub async fn get_video_info(&self, url: &str) -> Result<TikTokVideo> {
        let exe = Executor {
            executable_path: self.libs.youtube.clone(),
            timeout: Duration::from_secs(60),
            args: vec![url.to_string(), "--dump-json".to_string()],
        };
        let output = exe.execute().await?;

        Ok(serde_json::from_str(&output.stdout)?)
    }

    pub async fn download_video_from_url(&self, url: &str, output: &Path) -> Result<()> {
        let exe = Executor {
            executable_path: self.libs.youtube.clone(),
            timeout: Duration::from_secs(60),
            args: vec![
                url.to_string(),
                "-o".to_string(),
                output.to_string_lossy().into_owned(),
            ],
        };
        exe.execute().await?;
        Ok(())
    }
}

#[derive(Deserialize, Serialize)]
struct TikTokVideo {
    #[serde(rename = "fulltitle")]
    pub full_title: String,
}

#[derive(Deserialize, Serialize)]
struct NoviaVideoData {
    event_id: String,
    video: String,
}
