use crate::blossom::Blossom;
use crate::dvm::{build_status_for_job, create_dvms, DVMInput, DVMJobRequest};
use anyhow::{anyhow, Result};
use clap::builder::TypedValueParser;
use clap::Parser;
use log::{error, info, warn};
use nostr_sdk::async_utility::futures_util::SinkExt;
use nostr_sdk::prelude::{DataVendingMachineStatus, JobFeedbackData, StreamExt};
use nostr_sdk::{
    ClientBuilder, Event, EventBuilder, EventId, Filter, Keys, Kind, RelayPoolNotification, Tag,
    Timestamp,
};
use std::collections::HashMap;

mod blossom;
mod dvm;

#[derive(clap::Parser)]
struct Args {
    #[arg(long)]
    nsec: Option<String>,

    /// One of more relays to listen to events on
    #[arg(long)]
    relay: Option<Vec<String>>,

    /// One or more blossom servers to upload content to
    #[arg(long)]
    blossom: Vec<String>,
}

#[tokio::main]
async fn main() -> Result<()> {
    env_logger::init();

    let args = Args::parse();

    let keys = if let Some(keys) = args.nsec {
        Keys::parse(&keys)?
    } else {
        Keys::generate()
    };
    let client = ClientBuilder::new().signer(keys).build();

    if let Some(relays) = &args.relay {
        for relay in relays {
            info!("Connecting relay {}", &relay);
            client.add_relay(relay).await?;
        }
    } else {
        info!("Using default relays");
        client.add_relay("wss://relay.damus.io").await?;
        client.add_relay("wss://relay.primal.net").await?;
        client.add_relay("wss://relay.snort.social").await?;
        client.add_relay("wss://relay.nostr.band").await?;
        client.add_relay("wss://nos.lol").await?;
    }
    client.connect().await;

    // TODO: support multiple servers
    let blossom = Blossom::new(args.blossom.first().unwrap());
    let mut dvms = create_dvms(&client, &blossom);

    const JOB_REQUEST_KIND: u16 = 5_205;

    // https://github.com/teamnovia/novia?tab=readme-ov-file#dvm-archive-aka-download-request-kind-5205
    let sub = client
        .subscribe(
            Filter::new()
                .kind(Kind::from(JOB_REQUEST_KIND))
                .since(Timestamp::now()),
            None,
        )
        .await?;

    info!("Listening for jobs..");
    let mut rx = client.notifications();
    while let Ok(e) = rx.recv().await {
        match e {
            RelayPoolNotification::Event { event, .. } => match event.kind.as_u16() {
                JOB_REQUEST_KIND => match parse_job_request(&event) {
                    Ok(req) => {
                        for mut dvm in &mut dvms {
                            if let Err(e) = dvm.handle_request(req.clone()).await {
                                error!("Error handling job request: {}", e);

                                let data = build_status_for_job(
                                    &req,
                                    DataVendingMachineStatus::Error,
                                    Some(e.to_string().as_str()),
                                    None,
                                );
                                client.send_event_builder(data).await?;
                            }
                        }
                    }
                    Err(e) => warn!("Invalid job request: {:?}", e),
                },
                _ => {
                    warn!("Unknown event kind {}", event.kind);
                }
            },
            RelayPoolNotification::Message { .. } => {}
            RelayPoolNotification::Shutdown => {}
        }
    }

    client.unsubscribe(sub.val).await;
    Ok(())
}

fn parse_job_request(event: &Event) -> Result<DVMJobRequest> {
    let mut inputs = vec![];
    for i_tag in event
        .tags
        .iter()
        .filter(|t| t.kind().as_str() == "i")
        .map(|t| t.as_slice())
    {
        let input = match i_tag[2].as_str() {
            "url" => DVMInput::Url {
                url: if let Ok(u) = i_tag[1].parse() {
                    u
                } else {
                    warn!("Invalid url: {}", i_tag[1]);
                    continue;
                },
                relay: if i_tag.len() > 3 {
                    Some(i_tag[3].to_string())
                } else {
                    None
                },
                marker: if i_tag.len() > 4 {
                    Some(i_tag[4].to_string())
                } else {
                    None
                },
            },
            "event" => DVMInput::Event {
                event: if let Ok(t) = EventId::parse(&i_tag[1]) {
                    t
                } else {
                    warn!("Invalid event id: {}", i_tag[1]);
                    continue;
                },
                relay: if i_tag.len() > 3 {
                    Some(i_tag[3].to_string())
                } else {
                    None
                },
                marker: if i_tag.len() > 4 {
                    Some(i_tag[4].to_string())
                } else {
                    None
                },
            },
            "job" => DVMInput::Job {
                event: if let Ok(t) = EventId::parse(&i_tag[1]) {
                    t
                } else {
                    warn!("Invalid event id in job: {}", i_tag[1]);
                    continue;
                },
                relay: if i_tag.len() > 3 {
                    Some(i_tag[3].to_string())
                } else {
                    None
                },
                marker: if i_tag.len() > 4 {
                    Some(i_tag[4].to_string())
                } else {
                    None
                },
            },
            "text" => DVMInput::Text {
                data: i_tag[1].to_string(),
                relay: if i_tag.len() > 3 {
                    Some(i_tag[3].to_string())
                } else {
                    None
                },
                marker: if i_tag.len() > 4 {
                    Some(i_tag[4].to_string())
                } else {
                    None
                },
            },
            t => {
                warn!("unknown tag: {}", t);
                continue;
            }
        };
        inputs.push(input);
    }

    let params: HashMap<String, String> = event
        .tags
        .iter()
        .filter(|t| t.kind().as_str() == "param")
        .filter_map(|p| {
            let p = p.as_slice();
            if p.len() == 3 {
                Some((p[0].clone(), p[1].clone()))
            } else {
                warn!("Invalid param: {}", p.join(", "));
                None
            }
        })
        .collect();
    Ok(DVMJobRequest {
        event: event.clone(),
        inputs,
        output_type: event
            .tags
            .iter()
            .find(|t| t.kind().as_str() == "output")
            .and_then(|t| t.content())
            .map(|s| s.to_string()),
        params,
        bid: event
            .tags
            .iter()
            .find(|t| t.kind().as_str() == "bid")
            .and_then(|t| t.content())
            .and_then(|t| t.parse::<u64>().ok()),
        relays: event
            .tags
            .iter()
            .filter(|t| t.kind().as_str() == "relay")
            .map(|c| &c.as_slice()[1..])
            .flatten()
            .map(|s| s.to_string())
            .collect(),
    })
}
